# Copyright 2021 DB Netz AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Exceptions that may be raised by the PVMT module."""


class PropertyValueError(Exception):
    """Base class for all property-value related errors."""


class ScopeError(PropertyValueError):
    """Attempted to apply a PV group to an element outside its scope."""


class UndefinedKeyError(PropertyValueError, KeyError):
    """A key is attempted to be added which is not defined for the group."""


class CastingError(PropertyValueError):
    """A supplied value cannot be cast from or to the XML representation."""


class GroupNotAppliedError(PropertyValueError, KeyError):
    """The property value group has not been applied to this element."""
